package at.ac.uibk.gitsearch.service;


import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.service.impl.StatisticsServiceImpl;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Metrics;
import io.micrometer.core.instrument.Tags;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Service class for monitoring
 *
 * @author Johannes Kainz
 */
@Service
@Transactional
public class MonitoringService {
    private static final String USER_TAG = "classification";
    private static final String LANGUAGE_TAG = "language";
    private static final String EXERCISE_TAG = "exerciseID";

    private static final String USER_NAME = "number.users";
    private static final String EXERCISE_NAME = "number.exercises";
    private static final String LIKE_NAME = "number.likes";
    private static final String VIEW_NAME = "number.views";
    private static final String DOWNLOAD_NAME = "number.downloads";

    private final Logger log = LoggerFactory.getLogger(MonitoringService.class);

    private final UserService userService;
    private final SearchService searchService;
    private final LikesService likesService;
    private final StatisticsServiceImpl statisticsService;

    private final AtomicLong numberUsers;
    private final AtomicLong numberActiveUsers;
    private final Map<String, AtomicInteger> numberExercises;
    private final Map<String, AtomicLong> numberLikes;
    private final Map<String, AtomicLong> numberViews;
    private final Map<String, AtomicLong> numberDownloads;

    private final MeterRegistry meterRegistry;


    public MonitoringService(MeterRegistry meterRegistry, UserService userService, SearchService searchService, LikesService likesService, StatisticsServiceImpl statisticsService) {
        this.meterRegistry = meterRegistry;

        this.userService = userService;
        this.searchService = searchService;
        this.likesService = likesService;
        this.statisticsService = statisticsService;

        this.numberExercises = new HashMap<>();
        this.numberLikes = new HashMap<>();
        this.numberViews = new HashMap<>();
        this.numberDownloads = new HashMap<>();

        this.numberUsers = meterRegistry.gauge(USER_NAME, Tags.of(USER_TAG, "All"), new AtomicLong(this.userService.countAllUsers()));
        this.numberActiveUsers = meterRegistry.gauge(USER_NAME, Tags.of(USER_TAG, "Active"), new AtomicLong(this.userService.countAllActiveUsers()));

        try {
            final SearchResultsDTO searchResults = this.getAllSearchResults();
            final List<String> exerciseIDs = this.getExerciseIDs(searchResults.getSearchResult());

            this.setNumberExercises(exerciseIDs.size());
            this.setExercisesLanguageCount(searchResults);
            this.setExerciseStatistics(exerciseIDs);

        } catch (IOException exception){
            log.warn("An IOException occurred when searching for all exercises\n" + Arrays.toString(exception.getStackTrace()));
        }

    }


    /**
     * exports the monitoring relevant metrics regularly
     * <p></p>
     * This is scheduled to get fired every minute (for testing reasons)
     */
    @Scheduled(cron = "30 * * * * ?")
    public void exportMonitoringMetrics(){
        log.debug("Export monitoring relevant metrics");

        this.setNumberUsers();

        try{
            SearchResultsDTO searchResults = this.getAllSearchResults();
            List<String> exerciseIDs = this.getExerciseIDs(searchResults.getSearchResult());

            this.setNumberExercises(exerciseIDs.size());
            this.setExercisesLanguageCount(searchResults);
            this.setExerciseStatistics(exerciseIDs);


        } catch (IOException exception){
            log.warn("An IOException occurred\n" + Arrays.toString(exception.getStackTrace()));
        }
    }

    private void setNumberUsers(){
        log.debug("set the number of users");

        this.numberUsers.set(this.userService.countAllUsers());
        this.numberActiveUsers.set(this.userService.countAllActiveUsers());

    }


    /**
     * Sets likes, views and downloads for each given exerciseID
     *
     * @param exerciseIDs given exercise IDs
     */
    private void setExerciseStatistics(List<String> exerciseIDs){
        log.debug("set the exercise statistics for {}", exerciseIDs);

        exerciseIDs.forEach(exerciseID -> {
            this.setNumberLikesForExerciseID(exerciseID);

            final Optional<StatisticsDTO> statisticOptional = this.statisticsService.findOneByExerciseID(exerciseID);
            if( statisticOptional.isPresent()){
                StatisticsDTO statistic = statisticOptional.get();
                this.setNumberViewsForExerciseID(exerciseID, statistic.getViews());
                this.setNumberDownloadsForExerciseID(exerciseID, statistic.getDownloads());
            }



        });
    }


    /**
     * Sets the given value as number of all exercises
     *
     * @param value given value
     */
    private void setNumberExercises(int value){
        if(!this.numberExercises.containsKey("All")){
            this.numberExercises.put("All", meterRegistry.gauge(EXERCISE_NAME, Tags.of(LANGUAGE_TAG, "All"), new AtomicInteger(value)));
        }else{
            this.numberExercises.get("All").set(value);
        }
    }

    /**
     * Sets the per programming language exercise counters for given search results
     *
     * @param searchResults the search results
     */
    private void setExercisesLanguageCount(SearchResultsDTO searchResults){
        log.debug("set exercise language count");

        Map<String, Integer> languageCount = this.getLanguageCount(searchResults);
        for (String language: languageCount.keySet()) {
            if(!this.numberExercises.containsKey(language)){
                this.numberExercises
                    .put(language, Metrics.gauge(EXERCISE_NAME, Tags.of(LANGUAGE_TAG, language), new AtomicInteger(languageCount.get(language)) ));
            } else{
                this.numberExercises.get(language).set(languageCount.get(language));
            }
        }

    }

    /**
     * Gets a map containing the programming language counts for given search results
     *
     * @param searchResults the search results
     * @return a map containing the language counts
     */
    private Map<String, Integer> getLanguageCount(SearchResultsDTO searchResults){
        Map<String, Integer> languageCount = new HashMap<>();
        List<UserProvidedMetadataDTO> metaDataDTO = new ArrayList<>();

        List<SearchResultDTO> searchResultList = searchResults.getSearchResult();

        searchResultList.forEach(searchResult -> metaDataDTO.add(searchResult.getMetadata()));

        for (UserProvidedMetadataDTO metaData: metaDataDTO ) {
            String[] languageMetaData = metaData.getProgrammingLanguage();
            if (languageMetaData != null) {
                Arrays.stream(metaData.getProgrammingLanguage())
                    .forEach(language -> languageCount.put(language, languageCount.getOrDefault(language, 0) + 1 ));
            }else{
                languageCount.put("Unclassified", languageCount.getOrDefault("Unclassified",0)+1);
            }
        }

        return languageCount;
    }

    /**
     * Sets the number of likes for a given exercise ID
     *
     * @param exerciseID given exercise ID
     */
    private void setNumberLikesForExerciseID(String exerciseID){
        long numberLikes = this.likesService.findNumberOfLikesByExerciseID(exerciseID);

        if (!this.numberLikes.containsKey(exerciseID)){
            this.numberLikes.put(exerciseID, meterRegistry.gauge(LIKE_NAME, Tags.of(EXERCISE_TAG, exerciseID), new AtomicLong(numberLikes)));
        }else{
            this.numberLikes.get(exerciseID).set(numberLikes);
        }
    }

    /**
     * Sets the given number of views for a given exercise ID
     *
     *  @param exerciseID given exercise ID
     * @param views given number of views
     */
    private void setNumberViewsForExerciseID(String exerciseID, long views){
        if(!this.numberViews.containsKey(exerciseID)){
            this.numberViews.put(exerciseID, meterRegistry.gauge(VIEW_NAME, Tags.of(EXERCISE_TAG, exerciseID), new AtomicLong(views)));
        }else{
            this.numberViews.get(exerciseID).set(views);
        }
    }

    /**
     * Sets the  given number of downloads for a given exercise ID
     *
     *  @param exerciseID given exercise ID
     * @param downloads given number of downloads
     */
    private void setNumberDownloadsForExerciseID(String exerciseID, long downloads){
        if(!this.numberDownloads.containsKey(exerciseID)){
            this.numberDownloads.put(exerciseID, meterRegistry.gauge(DOWNLOAD_NAME, Tags.of(EXERCISE_TAG, exerciseID), new AtomicLong(downloads)));
        }else{
            this.numberDownloads.get(exerciseID).set(downloads);
        }
    }



    /**
     * Gets the search results of a search for all
     *
     * @return the search results DTO
     * @throws IOException exception for timeout etc.
     */
    private SearchResultsDTO getAllSearchResults() throws IOException {
        return this.searchService.searchAll();
    }

    /**
     * Gets the exercise ids from a search result list
     *
     * @param searchResultList the search result list
     * @return a list of exerciseIDs
     */
    private List<String> getExerciseIDs(List<SearchResultDTO> searchResultList){
        log.debug("Get the exercise IDs from a list of search results");

        List<String> exerciseIDs = new ArrayList<>();

        searchResultList.forEach(searchResult -> exerciseIDs.add(searchResult.getExerciseId()));
        return exerciseIDs;
    }

}
